# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Any
from .._typing_compat import Dict, Tuple, Type

from ._yadisk_object import YaDiskObject
from ._link_object import LinkObject

from ..types import OperationStatus, TimeoutParameter, Headers

class OperationStatusObject(YaDiskObject):
    status: OperationStatus

    def __init__(self,
                 operation_status: Optional[dict] = None,
                 yadisk: Optional[Any] = None) -> None:
        ...

class OperationLinkObject(LinkObject):
    ...

class SyncOperationLinkObject(OperationLinkObject):
    def get_status(
        self,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        requests_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        curl_options: Optional[Dict[int, Any]] = None,
        **kwargs
    ) -> OperationStatus: ...

    def wait(
        self,
        *,
        poll_interval: float = ...,
        poll_timeout: Optional[float] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        requests_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        curl_options: Optional[Dict[int, Any]] = None,
        **kwargs
    ) -> None:
        ...


class AsyncOperationLinkObject(OperationLinkObject):
    async def get_status(
        self,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> OperationStatus: ...

    async def wait(
        self,
        *,
        poll_interval: float = ...,
        poll_timeout: Optional[float] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> None:
        ...



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Any, Union, Literal, overload
from ._typing_compat import Dict, AsyncGenerator, Iterable, List, Tuple, Type

from .objects import (
    DeviceCodeObject, TokenObject, TokenRevokeStatusObject,
    DiskInfoObject, AsyncResourceObject, AsyncResourceLinkObject,
    AsyncOperationLinkObject, AsyncTrashResourceObject, AsyncPublicResourceObject,
    AsyncPublicResourcesListObject, AsyncPublicResourceLinkObject,
    ResourceUploadLinkObject
)

from .types import (
    AsyncFileOrPath, AsyncFileOrPathDestination, Headers, OperationStatus, AsyncSession,
    AsyncSessionName, AsyncSessionFactory, AsyncOpenFileCallback, TimeoutParameter
)

__all__ = ["AsyncClient"]

class AsyncClient:
    id: str
    secret: str
    token: str
    default_args: Dict[str, Any]
    session: AsyncSession
    open_file: AsyncOpenFileCallback

    synchronous = False

    def __init__(
        self,
        id:     str = "",
        secret: str = "",
        token:  str = "",
        *,
        default_args:    Optional[Dict[str, Any]] = None,
        session:         Optional[Union[AsyncSession, AsyncSessionName]] = None,
        open_file:       Optional[AsyncOpenFileCallback] = None,
        session_factory: Optional[AsyncSessionFactory] = None
    ) -> None:
        ...

    async def __aenter__(self): ...
    async def __aexit__(self, *args, **kwargs) -> None: ...
    async def close(self) -> None: ...

    def get_auth_url(
        self,
        type:                  Union[Literal["code"], Literal["token"]],
        device_id:             Optional[str] = None,
        device_name:           Optional[str] = None,
        redirect_uri:          Optional[str] = None,
        login_hint:            Optional[str] = None,
        scope:                 Optional[str] = None,
        optional_scope:        Optional[str] = None,
        force_confirm:         bool = True,
        state:                 Optional[str] = None,
        code_challenge:        Optional[str] = None,
        code_challenge_method: Optional[Union[Literal["plain"], Literal["S256"]]] = None,
        display:               None = None
    ) -> str:
        ...

    def get_code_url(
        self,
        device_id:             Optional[str] = None,
        device_name:           Optional[str] = None,
        redirect_uri:          Optional[str] = None,
        login_hint:            Optional[str] = None,
        scope:                 Optional[str] = None,
        optional_scope:        Optional[str] = None,
        force_confirm:         bool = True,
        state:                 Optional[str] = None,
        code_challenge:        Optional[str] = None,
        code_challenge_method: Optional[Union[Literal["plain"], Literal["S256"]]] = None,
        display:               None = None
    ) -> str: ...

    async def get_device_code(
        self,
        *,
        device_id: Optional[str] = None,
        device_name: Optional[str] = None,
        scope: Optional[str] = None,
        optional_scope: Optional[str] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> DeviceCodeObject: ...

    async def get_token(
        self,
        code: str,
        /,
        *,
        device_id: Optional[str] = None,
        device_name: Optional[str] = None,
        code_verifier: Optional[str] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> TokenObject:
        ...

    async def get_token_from_device_code(
        self,
        device_code: str,
        /,
        *,
        device_id: Optional[str] = None,
        device_name: Optional[str] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> TokenObject:
        ...

    async def refresh_token(
        self,
        refresh_token: str,
        /,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> TokenObject:
        ...

    async def revoke_token(
        self,
        token: Optional[str] = None,
        /,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> TokenRevokeStatusObject:
        ...

    async def check_token(
        self,
        token: Optional[str] = None,
        /,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> bool:
        ...

    async def get_disk_info(
        self,
        *,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> DiskInfoObject:
        ...

    async def get_meta(
        self,
        path: str,
        /,
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        preview_size: Optional[str] = None,
        preview_crop: Optional[bool] = None,
        sort: Optional[str] = None,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncResourceObject:
        ...

    async def exists(
        self,
        path: str,
        /,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> bool:
        ...

    async def get_type(
        self,
        path: str,
        /,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> str:
        ...

    async def is_file(
        self,
        path: str,
        /,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> bool:
        ...

    async def is_dir(
        self,
        path: str,
        /,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> bool:
        ...

    async def listdir(
        self,
        path: str,
        /,
        *,
        max_items: Optional[int] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        preview_size: Optional[str] = None,
        preview_crop: Optional[bool] = None,
        sort: Optional[str] = None,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncGenerator[AsyncResourceObject, None]:
        # This line here is needed so that the type checker knows that this is
        # an async generator, rather than a simple async function
        yield AsyncResourceObject()

    async def get_upload_link(
        self,
        path: str,
        /,
        *,
        overwrite: bool = False,
        spoof_user_agent: bool = True,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> str:
        ...

    async def get_upload_link_object(
        self,
        path: str,
        /,
        *,
        overwrite: bool = False,
        fields: Optional[Iterable[str]] = None,
        spoof_user_agent: bool = True,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ResourceUploadLinkObject:
        ...

    async def upload(
        self,
        file_or_path: AsyncFileOrPath,
        dst_path: str,
        /,
        *,
        overwrite: bool = False,
        spoof_user_agent: bool = True,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncResourceLinkObject:
        ...

    async def upload_by_link(
        self,
        file_or_path: AsyncFileOrPath,
        link: str,
        /,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> None:
        ...

    async def get_download_link(
        self,
        path: str,
        /,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> str:
        ...

    async def download(
        self,
        src_path: str,
        file_or_path: AsyncFileOrPathDestination,
        /,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncResourceLinkObject:
        ...

    async def download_by_link(
        self,
        link: str,
        file_or_path: AsyncFileOrPathDestination,
        /,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> None:
        ...

    @overload
    async def remove(
        self,
        path: str,
        /,
        *,
        wait: bool = True,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        md5: Optional[str] = None,
        force_async: Literal[True],
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncOperationLinkObject:
        ...

    @overload
    async def remove(
        self,
        path: str,
        /,
        *,
        wait: bool = True,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        md5: Optional[str] = None,
        force_async: bool = False,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> Optional[AsyncOperationLinkObject]:
        ...

    async def mkdir(
        self,
        path: str,
        /,
        *,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncResourceLinkObject:
        ...

    async def makedirs(
        self,
        path: str,
        /,
        *,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncResourceLinkObject:
        ...

    async def get_trash_meta(
        self,
        path: str,
        /,
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        preview_size: Optional[str] = None,
        preview_crop: Optional[bool] = None,
        sort: Optional[str] = None,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncTrashResourceObject:
        ...

    async def trash_exists(
        self,
        path: str,
        /,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> bool:
        ...

    @overload
    async def copy(
        self,
        src_path: str,
        dst_path: str,
        /,
        *,
        wait: bool = True,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        overwrite: bool = False,
        force_async: Literal[True],
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncOperationLinkObject:
        ...

    @overload
    async def copy(
        self,
        src_path: str,
        dst_path: str,
        /,
        *,
        wait: bool = True,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        overwrite: bool = False,
        force_async: bool = False,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> Union[AsyncResourceLinkObject, AsyncOperationLinkObject]:
        ...

    @overload
    async def restore_trash(
        self,
        path: str,
        /,
        dst_path: Optional[str] = None,
        *,
        wait: bool = True,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        overwrite: bool = False,
        force_async: Literal[True],
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncOperationLinkObject:
        ...

    @overload
    async def restore_trash(
        self,
        path: str,
        /,
        dst_path: Optional[str] = None,
        *,
        wait: bool = True,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        overwrite: bool = False,
        force_async: bool = False,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> Union[AsyncResourceLinkObject, AsyncOperationLinkObject]:
        ...

    @overload
    async def move(
        self,
        src_path: str,
        dst_path: str,
        /,
        *,
        wait: bool = True,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        overwrite: bool = False,
        force_async: Literal[True],
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncOperationLinkObject:
        ...

    @overload
    async def move(
        self,
        src_path: str,
        dst_path: str,
        /,
        *,
        wait: bool = True,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        overwrite: bool = False,
        force_async: bool = False,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> Union[AsyncResourceLinkObject, AsyncOperationLinkObject]:
        ...

    @overload
    async def rename(
        self,
        src_path: str,
        new_name: str,
        /,
        *,
        wait: bool = True,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        overwrite: bool = False,
        force_async: Literal[True],
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncOperationLinkObject:
        ...

    @overload
    async def rename(
        self,
        src_path: str,
        new_name: str,
        /,
        *,
        wait: bool = True,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        overwrite: bool = False,
        force_async: bool = False,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> Union[AsyncResourceLinkObject, AsyncOperationLinkObject]:
        ...

    @overload
    async def remove_trash(
        self,
        path: str,
        /,
        *,
        wait: bool = True,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        force_async: Literal[True],
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncOperationLinkObject:
        ...

    @overload
    async def remove_trash(
        self,
        path: str,
        /,
        *,
        wait: bool = True,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        force_async: bool = False,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> Optional[AsyncOperationLinkObject]:
        ...

    async def publish(
        self,
        path: str,
        /,
        *,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncResourceLinkObject:
        ...

    async def unpublish(
        self,
        path: str,
        /,
        *,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncResourceLinkObject:
        ...

    @overload
    async def save_to_disk(
        self,
        public_key: str,
        /,
        *,
        wait: bool = True,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        name: Optional[str] = None,
        path: Optional[str] = None,
        save_path: Optional[str] = None,
        force_async: Literal[True],
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncOperationLinkObject:
        ...

    @overload
    async def save_to_disk(
        self,
        public_key: str,
        /,
        *,
        wait: bool = True,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        name: Optional[str] = None,
        path: Optional[str] = None,
        save_path: Optional[str] = None,
        force_async: bool = False,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> Union[AsyncResourceLinkObject, AsyncOperationLinkObject]:
        ...

    async def get_public_meta(
        self,
        public_key: str,
        /,
        *,
        path: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        preview_size: Optional[str] = None,
        preview_crop: Optional[bool] = None,
        sort: Optional[str] = None,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncPublicResourceObject:
        ...

    async def public_exists(
        self,
        public_key: str,
        /,
        *,
        path: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        preview_size: Optional[str] = None,
        preview_crop: Optional[bool] = None,
        sort: Optional[str] = None,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> bool:
        ...

    async def public_listdir(
        self,
        public_key: str,
        /,
        *,
        path: Optional[str] = None,
        max_items: Optional[int] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        preview_size: Optional[str] = None,
        preview_crop: Optional[bool] = None,
        sort: Optional[str] = None,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncGenerator[AsyncPublicResourceObject, None]:
        # This line here is needed so that the type checker knows that this is
        # an async generator, rather than a simple async function
        yield AsyncPublicResourceObject()

    async def get_public_type(
        self,
        public_key: str,
        /,
        *,
        path: Optional[str] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> str:
        ...

    async def is_public_dir(
        self,
        public_key: str,
        /,
        *,
        path: Optional[str] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> bool:
        ...

    async def is_public_file(
        self,
        public_key: str,
        /,
        *,
        path: Optional[str] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> bool:
        ...

    async def trash_listdir(
        self,
        path: str,
        /,
        *,
        limit: Optional[int] = None,
        max_items: Optional[int] = None,
        offset: Optional[int] = None,
        preview_size: Optional[str] = None,
        preview_crop: Optional[bool] = None,
        sort: Optional[str] = None,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncGenerator[AsyncTrashResourceObject, None]:
        # This line here is needed so that the type checker knows that this is
        # an async generator, rather than a simple async function
        yield AsyncTrashResourceObject()

    async def get_trash_type(
        self,
        path: str,
        /,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> str:
        ...

    async def is_trash_dir(
        self,
        path: str,
        /,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> bool:
        ...

    async def is_trash_file(
        self,
        path: str,
        /,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> bool:
        ...

    async def get_public_resources(
        self,
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        preview_size: Optional[str] = None,
        preview_crop: Optional[bool] = None,
        type: Optional[Union[Literal["file"], Literal["dir"]]] = None,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncPublicResourcesListObject:
        ...

    async def get_all_public_resources(
        self,
        *,
        max_items: Optional[int] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        preview_size: Optional[str] = None,
        preview_crop: Optional[bool] = None,
        type: Optional[Union[Literal["file"], Literal["dir"]]] = None,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncGenerator[AsyncPublicResourceObject, None]:
        yield AsyncPublicResourceObject()

    async def patch(
        self,
        path: str,
        properties: dict,
        /,
        *,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncResourceObject:
        ...

    async def get_files(
        self,
        *,
        max_items: Optional[int] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        media_type: Optional[Union[str, Iterable[str]]] = None,
        sort: Optional[str] = None,
        preview_size: Optional[str] = None,
        preview_crop: Optional[bool] = None,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncGenerator[AsyncResourceObject, None]:
        # This line here is needed so that the type checker knows that this is
        # an async generator, rather than a simple async function
        yield AsyncResourceObject()

    async def get_last_uploaded(
        self,
        *,
        limit: Optional[int] = None,
        media_type: Optional[Union[str, Iterable[str]]] = None,
        preview_size: Optional[str] = None,
        preview_crop: Optional[bool] = None,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> List[AsyncResourceObject]: ...

    async def upload_url(
        self,
        url: str,
        path: str,
        /,
        *,
        wait: bool = True,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        disable_redirects: bool = False,
        fields: Optional[Iterable[str]] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncOperationLinkObject:
        ...

    async def get_public_download_link(
        self,
        public_key: str,
        /,
        *,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> str:
        ...

    async def download_public(
        self,
        public_key: str,
        file_or_path: AsyncFileOrPathDestination,
        /,
        *,
        path: Optional[str] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> AsyncPublicResourceLinkObject:
        ...

    async def get_operation_status(
        self,
        operation_id: str,
        /,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> OperationStatus:
        ...

    async def wait_for_operation(
        self,
        operation_id: str,
        /,
        *,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        headers: Optional[Headers] = None,
        timeout: TimeoutParameter = ...,
        n_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        retry_on: Tuple[Type[Exception], ...] = tuple(),
        aiohttp_args: Optional[Dict[str, Any]] = None,
        httpx_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> None:
        ...



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from .utils import CaseInsensitiveDict

from ._typing_compat import Dict, Generator
from .exceptions import InvalidResponseError
from .objects import ResourceObject, LinkObject

from typing import Any, AnyStr, IO, Optional

__all__ = [
    "_add_authorization_header",
    "_add_spoof_user_agent_header",
    "_apply_default_args",
    "_filter_request_kwargs",
    "_read_file_as_generator",
    "_set_authorization_header",
    "_validate_get_type_response",
    "_validate_link_response",
    "_validate_listdir_response"
]


def _apply_default_args(args: Dict[str, Any], default_args: Dict[str, Any]) -> None:
    new_args = dict(default_args)
    new_args.update(args)
    args.clear()
    args.update(new_args)


def _filter_request_kwargs(kwargs: Dict[str, Any]) -> None:
    # Remove some of the yadisk-specific arguments from kwargs
    keys_to_remove = ("n_retries", "retry_interval", "retry_on", "fields", "overwrite", "path")

    for key in keys_to_remove:
        kwargs.pop(key, None)


def _read_file_as_generator(input_file: IO[AnyStr]) -> Generator[AnyStr, None, None]:
    chunk_size = 8192

    while chunk := input_file.read(chunk_size):
        yield chunk


def _set_authorization_header(
    kwargs: Dict[str, Any],
    new_token: Optional[str] = None
) -> None:
    headers = CaseInsensitiveDict(kwargs.get("headers") or {})

    if new_token:
        headers["Authorization"] = f"OAuth {new_token}"
    else:
        headers.pop("Authorization", None)

    kwargs["headers"] = headers


def _add_authorization_header(
    kwargs: Dict[str, Any],
    new_token: Optional[str] = None
) -> None:
    headers = CaseInsensitiveDict(kwargs.get("headers") or {})

    if "Authorization" in headers:
        return

    if new_token:
        headers["Authorization"] = f"OAuth {new_token}"
    else:
        headers["Authorization"] = ""

    kwargs["headers"] = headers


def _add_spoof_user_agent_header(kwargs: Dict[str, Any]) -> None:
    headers = kwargs.setdefault("headers", {})
    headers["User-Agent"] = 'Yandex.Disk {"os":"windows"}'


def _validate_listdir_response(response: ResourceObject) -> ResourceObject:
    if response.type == "file":
        return response

    if response.embedded is None:
        raise InvalidResponseError("Response did not contain _embedded field")

    if (response.type is None or response.embedded.items is None or
            response.embedded.offset is None or response.embedded.limit is None or
            response.embedded.total is None):
        raise InvalidResponseError("Response did not contain key field")

    return response

def _validate_link_response(response: LinkObject) -> LinkObject:
    if not response.href:
        raise InvalidResponseError("Response did not contain the link")

    return response


def _validate_get_type_response(response: ResourceObject) -> ResourceObject:
    if response.type is None:
        raise InvalidResponseError("Response did not contain the type field")

    return response



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from typing import Any

from .._async_session import AsyncSession, AsyncResponse
from ..types import JSON, AsyncConsumeCallback, HTTPMethod
from .._common import is_async_func

from ._httpx_common import *

import httpx

__all__ = ["AsyncHTTPXSession"]


class AsyncHTTPXResponse(AsyncResponse):
    def __init__(self, response: httpx.Response):
        super().__init__()

        self._response = response
        self.status = response.status_code

    async def json(self) -> JSON:
        try:
            await self._response.aread()
        except httpx.HTTPError as e:
            raise convert_httpx_exception(e) from e
        except httpx.StreamConsumed as e:
            raise ValueError(f"Could not parse JSON: {e}") from e

        return self._response.json()

    async def download(self, consume_callback: AsyncConsumeCallback) -> None:
        callback: Any = consume_callback

        try:
            if is_async_func(consume_callback):
                async for chunk in self._response.aiter_bytes(8192):
                    await callback(chunk)
            else:
                async for chunk in self._response.aiter_bytes(8192):
                    callback(chunk)
        except httpx.HTTPError as e:
            raise convert_httpx_exception(e) from e

    async def close(self) -> None:
        await self._response.aclose()


class AsyncHTTPXSession(AsyncSession):
    """
        .. _httpx: https://pypi.org/project/httpx

        :any:`AsyncSession` implementation using the `httpx`_ library.

        .. _httpx.AsyncClient: https://www.python-httpx.org/api/#asyncclient

        All arguments passed in the constructor are directly forwared to `httpx.AsyncClient`_.

        :ivar httpx_client: underlying instance of `httpx.AsyncClient`_

        To pass `httpx`-specific arguments from :any:`AsyncClient` use :code:`httpx_args` keyword argument.

        Usage example:

        .. code:: python

           import yadisk

           async def main():
               async with yadisk.AsyncClient(..., session="httpx") as client:
                   await client.get_meta(
                       "/my_file.txt",
                       n_retries=5,
                       httpx_args={
                           "proxy": "http://localhost:11234",
                           "verify": False,
                           "max_redirects": 10
                       }
                    )
    """

    def __init__(self, *args, **kwargs):
        self._session = httpx.AsyncClient(*args, **kwargs)
        self._session.follow_redirects = True

    @property
    def httpx_session(self) -> httpx.AsyncClient:
        return self._session

    async def send_request(self, method: HTTPMethod, url: str, **kwargs) -> AsyncResponse:
        request_kwargs, send_kwargs = convert_args_for_httpx(self._session, kwargs)

        try:
            request = self._session.build_request(method, url, **request_kwargs)
            return AsyncHTTPXResponse(await self._session.send(request, **send_kwargs))
        except httpx.HTTPError as e:
            raise convert_httpx_exception(e) from e

    async def close(self) -> None:
        await self._session.aclose()



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

import asyncio
from ..exceptions import InvalidResponseError

from ..utils import auto_retry, async_auto_retry, CaseInsensitiveDict
from .. import settings
from .._common import is_default_timeout

from typing import Any, Optional, Union, Type, TypeVar, TYPE_CHECKING
from .._typing_compat import Set, Dict, Tuple, Callable, Awaitable
import json

if TYPE_CHECKING:  # pragma: no cover
    from .._session import Session
    from .._async_session import AsyncSession
    from ..types import AnySession, HTTPMethod, JSON, TimeoutParameter
    from .._client import Client
    from .._async_client import AsyncClient

__all__ = ["APIRequest"]


class APIRequest(object):
    """
        Base class for all API requests.

        :param session: an instance of :any:`Session`
        :param timeout: `float` or `tuple`, request timeout
        :param headers: `dict` or `None`, additional request headers
        :param n_retries: `int`, maximum number of retries
        :param retry_interval: delay between retries in seconds
        :param retry_on: `tuple`, additional exception classes to retry on
        :param kwargs: other arguments for :any:`Session.send_request`

        :ivar base_url: `str`, base URL for sending the request
        :ivar url: `str`, request URL
        :ivar path: `str`, URL path for the request
        :ivar method: `str`, request method
        :ivar content_type: `str`, Content-Type header ("application/x-www-form-urlencoded" by default)
        :ivar timeout: `float` or `tuple`, request timeout
        :ivar n_retries: `int`, maximum number of retries
        :ivar success_codes: `list`-like, list of response codes that indicate request's success
        :ivar retry_interval: `float`, delay between retries in seconds
        :ivar retry_on: `tuple`, additional exception classes to retry on
    """

    base_url: str = ""
    url: str = ""
    path: str = ""
    method: Optional["HTTPMethod"] = None
    content_type: str = "application/x-www-form-urlencoded"

    timeout: "TimeoutParameter"
    n_retries: Optional[int] = None
    success_codes: Set[int] = {200}
    retry_interval: Optional[Union[int, float]] = None

    data: Union[Dict, bytes]
    params: Dict[str, Any]
    send_kwargs: Dict[str, Any]
    retry_on: Tuple[Type[Exception], ...] = tuple()

    session: Any

    T = TypeVar("T")

    def __init__(self, session: "AnySession", **kwargs):
        base_url = self.base_url or settings.BASE_API_URL
        n_retries = kwargs.pop("n_retries", None)
        retry_interval = kwargs.pop("retry_interval", None)
        headers = kwargs.pop("headers", {})
        retry_on = kwargs.pop("retry_on", self.retry_on)

        if headers is None:
            headers = {}

        timeout = kwargs.get("timeout", ...)

        if is_default_timeout(timeout):
            timeout = settings.DEFAULT_TIMEOUT

        kwargs["timeout"] = timeout

        if n_retries is None:
            n_retries = self.n_retries
        if n_retries is None:
            n_retries = settings.DEFAULT_N_RETRIES

        if retry_interval is None:
            retry_interval = self.retry_interval
        if retry_interval is None:
            retry_interval = settings.DEFAULT_RETRY_INTERVAL

        self.session = session
        self.send_kwargs = kwargs
        self.base_url = base_url
        self.timeout = timeout
        self.n_retries = n_retries
        self.retry_interval = retry_interval
        self.headers = headers
        self.data = {}
        self.content = None
        self.params = {}
        self.retry_on = retry_on

        if not self.url:
            self.url = f"{self.base_url}/{self.path.lstrip('/')}"

    def _prepare_send_args(self) -> Dict[str, Any]:
        headers = CaseInsensitiveDict()
        headers["Content-Type"] = self.content_type
        headers.update(self.headers)

        if self.data:
            if isinstance(self.data, Dict):
                data = json.dumps(self.data).encode("utf8")
            else:
                data = self.data
        else:
            data = None

        kwargs = dict(self.send_kwargs)
        kwargs.update({"headers": headers,
                       "data":    data,
                       "params":  self.params})

        return kwargs

    def _attempt(
        self,
        yadisk: Optional["Client"],
        then: Callable[[Any], Any]
    ) -> Any:
        assert self.method is not None
        assert self.url

        kwargs = self._prepare_send_args()

        session: "Session" = self.session
        response = session.send_request(self.method, self.url, **kwargs)

        json: JSON = None

        if response.status == 0:
            # Request has not been sent yet
            # This can happen with pycurl with stream=True
            try:
                json = response.json()
            except ValueError:
                pass

            json_already_parsed = True
        else:
            json_already_parsed = False

        success = response.status in self.success_codes

        if not success:
            raise response.get_exception()

        if not json_already_parsed:
            try:
                json = response.json()
            except ValueError:
                pass

        try:
            result = self.process_json(json, yadisk=yadisk)
        except ValueError as e:
            raise InvalidResponseError(f"Server returned invalid response: {e}") from e

        return then(result)

    async def _async_attempt(
        self,
        yadisk: Optional["AsyncClient"],
        then: Union[Callable[[Any], Any], Callable[[Any], Awaitable[Any]]]
    ) -> Any:
        assert self.method is not None
        assert self.url

        kwargs = self._prepare_send_args()

        session: "AsyncSession" = self.session

        response = await session.send_request(self.method, self.url, **kwargs)

        success = response.status in self.success_codes

        if not success:
            raise await response.get_exception()

        try:
            json = await response.json()
        except ValueError:
            json = None

        try:
            result = self.process_json(json, yadisk=yadisk)
        except ValueError as e:
            raise InvalidResponseError(f"Server returned invalid response: {e}") from e

        if asyncio.iscoroutinefunction(then):
            return await then(result)
        else:
            return then(result)

    def send(
        self,
        yadisk: Optional["Client"],
        then: Optional[Callable[[Any], Any]] = None
    ) -> Any:
        """
            Actually send the request

            :param yadisk: :any:`Client` instance that will be passed to :any:`process_json()`
            :param then: function that will be called at the end of the attempt

            :returns: :any:`Response` (`self.response`)
        """

        settings.logger.info(f"sending APIRequest {self.__class__.__name__}, {self.method} {self.url}")

        return auto_retry(
            self._attempt,
            self.n_retries,
            self.retry_interval,
            args=(yadisk, then or (lambda x: x)),
            retry_on=self.retry_on
        )

    async def asend(
        self,
        yadisk: Optional["AsyncClient"],
        then: Optional[Union[Callable[[Any], Any], Callable[[Any], Awaitable[Any]]]] = None
    ) -> Any:
        """
            Actually send the request

            :param yadisk: :any:`AsyncClient` instance that will be passed to :any:`process_json()`
            :param then: function that will be called at the end of the attempt

            :returns: :any:`AsyncResponse` (`self.response`)
        """

        settings.logger.info(f"sending APIRequest {self.__class__.__name__}, {self.method} {self.url}")

        return await async_auto_retry(
            self._async_attempt,
            self.n_retries,
            self.retry_interval,
            args=(yadisk, then or (lambda x: x)),
            retry_on=self.retry_on
        )

    def process_json(self, js: "JSON", **kwargs) -> Any:
        """
            Process the JSON response.

            :param js: `dict` or `None`, JSON response
            :param kwargs: extra arguments (optional)

            :returns: processed response, can be anything
        """

        raise NotImplementedError



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

import asyncio
import datetime
import inspect

from ._typing_compat import Callable, List
from . import settings

from typing import Optional, TypeVar, Any, Union

from .types import TimeoutParameter, Tuple

__all__ = [
    "bool_or_error",
    "dict_or_error",
    "ensure_path_has_schema",
    "float_or_error",
    "int_or_error",
    "is_default_timeout",
    "is_operation_link",
    "is_public_resource_link",
    "is_resource_link",
    "remove_path_schema",
    "str_or_dict_or_error",
    "str_or_error",
    "typed_list",
    "yandex_date"
]

T = TypeVar("T", bound=Callable)


def typed_list(datatype: T) -> Callable[[Optional[List]], List[T]]:
    def list_factory(iterable: Optional[List] = None) -> List[T]:
        if iterable is None:
            return []

        if not isinstance(iterable, list):
            raise ValueError(f"Expected a list, got {type(iterable)}")

        return [datatype(i) for i in iterable]

    return list_factory


def int_or_error(x: Any) -> int:
    if not isinstance(x, int):
        raise ValueError(f"{repr(x)} is not an integer")

    return x


def float_or_error(x: Any) -> float:
    if not isinstance(x, (float, int)):
        raise ValueError(f"{repr(x)} is not a float")

    return x


def str_or_error(x: Any) -> str:
    if not isinstance(x, str):
        raise ValueError(f"{repr(x)} is not a string")

    return x


def bool_or_error(x: Any) -> bool:
    if not isinstance(x, bool):
        raise ValueError(f"{repr(x)} is not a boolean value")

    return x


def dict_or_error(x: Any) -> dict:
    if not isinstance(x, dict):
        raise ValueError(f"{repr(x)} is not a dict")

    return x


def str_or_dict_or_error(x: Any) -> Union[str, dict]:
    if not isinstance(x, (str, dict)):
        raise ValueError(f"{repr(x)} is not a string nor a dict")

    return x


def yandex_date(string: str) -> datetime.datetime:
    return datetime.datetime.strptime(string[:-3] + string[-2:], "%Y-%m-%dT%H:%M:%S%z")


def _is_endpoint_link(link: str, base_endpoint_url: str) -> bool:
    link_schema, _, link = link.partition("://")
    endpoint_schema, _, base_endpoint_url = base_endpoint_url.partition("://")

    if link_schema not in ("http", "https") or endpoint_schema not in ("http", "https"):
        return False

    if not base_endpoint_url.endswith("/") and not base_endpoint_url.endswith("?"):
        base_endpoint_url += "/"

    return link.startswith(base_endpoint_url)


def is_operation_link(link: str) -> bool:
    return _is_endpoint_link(link, f"{settings.BASE_API_URL}/v1/disk/operations/")


def is_resource_link(url: str) -> bool:
    return _is_endpoint_link(url, f"{settings.BASE_API_URL}/v1/disk/resources?")


def is_public_resource_link(url: str) -> bool:
    return _is_endpoint_link(url, f"{settings.BASE_API_URL}/v1/disk/public/resources?")


KNOWN_SCHEMAS = ("disk:", "trash:", "app:", "photounlim:")


def ensure_path_has_schema(path: str, default_schema: str = "disk") -> str:
    # Modifies path to always have a schema (disk:/, trash:/ or app:/).
    # Without the schema Yandex.Disk won't let you upload filenames with the ':' character.
    # See https://github.com/ivknv/yadisk/issues/26 for more details

    if path in KNOWN_SCHEMAS:
        return default_schema + ":/" + path

    if path.startswith("/"):
        return default_schema + ":" + path

    if any(path.startswith(schema + "/") for schema in KNOWN_SCHEMAS):
        return path

    return default_schema + ":/" + path


def remove_path_schema(path: str) -> Tuple[str, str]:
    """
        Remove schema from path.

        :param path: `str`, path to remove the schema from

        :returns: `tuple[str, str]`, removed schema (without `:/`) and the path without it
    """

    if path.startswith("/") or path in KNOWN_SCHEMAS:
        return "", path

    if any(path.startswith(schema + "/") for schema in KNOWN_SCHEMAS):
        schema, sep, path = path.partition(":/")

        return schema, path

    return "", path


def is_async_func(func: Any) -> bool:
    return inspect.isgeneratorfunction(func) or asyncio.iscoroutinefunction(func)

def is_default_timeout(timeout: TimeoutParameter) -> bool:
    return timeout is ...



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from functools import partial
from ._yadisk_object import YaDiskObject
from .._typing_compat import Dict
from .._common import str_or_error, bool_or_error, int_or_error, yandex_date

from typing import Any, Optional, NoReturn, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    import datetime

__all__ = ["DiskInfoObject", "SystemFoldersObject", "UserObject", "UserPublicInfoObject"]


class DiskInfoObject(YaDiskObject):
    """
        Disk information object.

        :param disk_info: `dict` or `None`
        :param yadisk: :any:`YaDisk` or `None`, `YaDisk` object

        :ivar paid_max_file_size: `int`, maximum supported file size for a paid account (bytes)
        :ivar max_file_size: `int`, maximum supported file size (bytes)
        :ivar total_space: `int`, total disk size (bytes)
        :ivar trash_size: `int`, amount of space used by trash (bytes), part of `used_space`
        :ivar used_space: `int`, amount of space used (bytes)
        :ivar is_paid: `bool`, tells if the account is paid or not
        :ivar is_idm_managed_folder_address_access: `bool`, not clear what this is for
        :ivar reg_time: :any:`datetime.datetime`, Disk registration date
        :ivar system_folders: :any:`SystemFoldersObject`, paths to the system folders
        :ivar user: :any:`UserObject`, owner of the disk
        :ivar is_idm_managed_public_access: `bool`, not clear what this is for
        :ivar unlimited_autoupload_enabled: `bool`, tells whether unlimited
                                             autoupload from mobile devices is enabled
        :ivar revision: `int`, current revision of Yandex.Disk
        :ivar photounlim_size: `int`, total file size in unlimited photos
        :ivar will_be_overdrawn: `bool`, tells if the user will be in overdraft upon
                                 reaching `free_photounlim_end_date`
        :ivar free_photounlim_end_date: `int`, timestamp in ms of expiration date of
                                        unlimited photo upload
        :ivar payment_flow: `bool`, tells if the user is involved in `payment_flow`
    """

    paid_max_file_size:                   Optional[int]
    max_file_size:                        Optional[int]
    total_space:                          Optional[int]
    trash_size:                           Optional[int]
    used_space:                           Optional[int]
    is_paid:                              Optional[bool]
    is_idm_managed_folder_address_access: Optional[bool]
    reg_time:                             Optional["datetime.datetime"]
    system_folders:                       "SystemFoldersObject"
    user:                                 "UserObject"
    is_idm_managed_public_access:         Optional[bool]
    unlimited_autoupload_enabled:         Optional[bool]
    revision:                             Optional[int]
    photounlim_size:                      Optional[int]
    will_be_overdrawn:                    Optional[bool]
    free_photounlim_end_date:             Optional[int]
    payment_flow:                         Optional[bool]

    def __init__(self, disk_info: Optional[Dict] = None, yadisk: Optional[Any] = None) -> None:
        YaDiskObject.__init__(
            self,
            {"paid_max_file_size":                   int_or_error,
             "max_file_size":                        int_or_error,
             "total_space":                          int_or_error,
             "trash_size":                           int_or_error,
             "used_space":                           int_or_error,
             "is_paid":                              bool_or_error,
             "is_idm_managed_folder_address_access": bool_or_error,
             "reg_time":                             yandex_date,
             "system_folders":                       partial(SystemFoldersObject, yadisk=yadisk),
             "user":                                 partial(UserObject, yadisk=yadisk),
             "is_idm_managed_public_access":         bool_or_error,
             "unlimited_autoupload_enabled":         bool_or_error,
             "revision":                             int_or_error,
             "photounlim_size":                      int_or_error,
             "will_be_overdrawn":                    bool_or_error,
             "free_photounlim_end_date":             int_or_error,
             "payment_flow":                         bool_or_error
            },
            yadisk
        )

        self.import_fields(disk_info)


class SystemFoldersObject(YaDiskObject):
    """
        Object, containing paths to system folders.

        :param system_folders: `dict` or `None`
        :param yadisk: :any:`YaDisk` or `None`, `YaDisk` object

        :ivar odnoklassniki: `str`, path to the Odnoklassniki folder
        :ivar google: `str`, path to the Google+ folder
        :ivar instagram: `str`, path to the Instagram folder
        :ivar vkontakte: `str`, path to the VKontakte folder
        :ivar attach: `str`, path to the mail attachments folder
        :ivar mailru: `str`, path to the My World folder
        :ivar downloads: `str`, path to the Downloads folder
        :ivar applications: `str` path to the Applications folder
        :ivar facebook: `str`, path to the Facebook folder
        :ivar social: `str`, path to the social networks folder
        :ivar messenger: `str`, path to the Messenger Files folder
        :ivar calendar: `str`, path to the Meeting Materials folder
        :ivar photostream: `str`, path to the camera folder
        :ivar screenshots: `str`, path to the screenshot folder
        :ivar scans: `str`, path to the Scans folder
    """

    odnoklassniki: Optional[str]
    google:        Optional[str]
    instagram:     Optional[str]
    vkontakte:     Optional[str]
    attach:        Optional[str]
    mailru:        Optional[str]
    downloads:     Optional[str]
    applications:  Optional[str]
    facebook:      Optional[str]
    social:        Optional[str]
    messenger:     Optional[str]
    calendar:      Optional[str]
    photostream:   Optional[str]
    screenshots:   Optional[str]
    scans:         Optional[str]

    def __init__(
        self,
        system_folders: Optional[Dict] = None,
        yadisk: Optional[Any] = None
    ) -> None:
        YaDiskObject.__init__(
            self,
            {"odnoklassniki": str_or_error,
             "google":        str_or_error,
             "instagram":     str_or_error,
             "vkontakte":     str_or_error,
             "attach":        str_or_error,
             "mailru":        str_or_error,
             "downloads":     str_or_error,
             "applications":  str_or_error,
             "facebook":      str_or_error,
             "social":        str_or_error,
             "messenger":     str_or_error,
             "calendar":      str_or_error,
             "photostream":   str_or_error,
             "screenshots":   str_or_error,
             "scans":         str_or_error},
            yadisk)

        self.import_fields(system_folders)


class UserObject(YaDiskObject):
    """
        User object.

        :param user: `dict` or `None`
        :param yadisk: :any:`YaDisk` or `None`, `YaDisk` object

        :ivar reg_time: :any:`datetime.datetime`, Disk registration date
        :ivar display_name: `str`, user's display name
        :ivar uid: `str`, user's UID
        :ivar country: `str`, user's country
        :ivar is_child: `bool`, tells whether it's a child account
        :ivar login: `str`, user's login
    """

    reg_time:     Optional["datetime.datetime"]
    display_name: Optional[str]
    uid:          Optional[str]
    country:      Optional[str]
    is_child:     Optional[bool]
    login:        Optional[str]

    def __init__(
        self,
        user: Optional[Dict] = None,
        yadisk: Optional[Any] = None
    ) -> None:
        YaDiskObject.__init__(
            self,
            {
                "reg_time":     yandex_date,
                "display_name": str_or_error,
                "uid":          str_or_error,
                "country":      str_or_error,
                "is_child":     bool_or_error,
                "login":        str_or_error
            },
            yadisk
        )

        self.import_fields(user)


class UserPublicInfoObject(UserObject):
    """
        Public user information object.
        Inherits from :any:`UserObject` for compatibility.

        :param public_user_info: `dict` or `None`
        :param yadisk: :any:`YaDisk` or `None`, `YaDisk` object

        :ivar login: `str`, user's login
        :ivar display_name: `str`, user's display name
        :ivar uid: `str`, user's UID
    """

    country: NoReturn  # pyright: ignore[reportIncompatibleVariableOverride]

    def __init__(
        self,
        public_user_info: Optional[Dict] = None,
        yadisk: Optional[Any] = None
    ) -> None:
        UserObject.__init__(self, None, yadisk)
        self.remove_field("country")
        self.import_fields(public_user_info)



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

__all__ = [
    "AsyncOperationFailedError",
    "AsyncOperationPollingTimeoutError",
    "AuthorizationPendingError",
    "BadGatewayError",
    "BadRequestError",
    "BadVerificationCodeError",
    "ConflictError",
    "DirectoryExistsError",
    "FieldValidationError",
    "ForbiddenError",
    "GatewayTimeoutError",
    "GoneError",
    "InsufficientStorageError",
    "InternalServerError",
    "InvalidClientError",
    "InvalidGrantError",
    "InvalidResponseError",
    "LockedError",
    "MD5DifferError",
    "NotAcceptableError",
    "NotFoundError",
    "OperationNotFoundError",
    "ParentNotFoundError",
    "PathExistsError",
    "PathNotFoundError",
    "PayloadTooLargeError",
    "RequestError",
    "RequestTimeoutError",
    "ResourceDownloadLimitExceededError",
    "ResourceIsLockedError",
    "RetriableYaDiskError",
    "TooManyRedirectsError",
    "TooManyRequestsError",
    "UnauthorizedError",
    "UnavailableError",
    "UnknownYaDiskError",
    "UnsupportedMediaError",
    "UnsupportedTokenTypeError",
    "UploadTrafficLimitExceededError",
    "WrongResourceTypeError",
    "YaDiskConnectionError",
    "YaDiskError"
]

from .types import AnyResponse
from typing import Optional


class YaDiskError(Exception):
    """
        Base class for all exceptions in this library.

        :ivar error_type: `str`, unique error code as returned by API
        :ivar response: an instance of :any:`Response` or :any:`AsyncResponse`
        :ivar disable_retry: `bool`, if set to :code:`True`, exception will not
                             trigger a retry in :any:`utils.auto_retry()`

        :param error_type: `str`, unique error code as returned by API
        :param msg: `str`, exception message
        :param response: an instance of :any:`Response` or :any:`AsyncResponse`
        :param disable_retry: `bool`, if set to :code:`True`, exception will not
                              trigger a retry in :any:`utils.auto_retry()`
    """

    error_type: Optional[str]
    response: Optional[AnyResponse]
    disable_retry: bool

    def __init__(
        self,
        error_type: Optional[str] = None,
        msg: str = "",
        response: Optional[AnyResponse] = None,
        disable_retry: bool = False
    ) -> None:
        Exception.__init__(self, msg)

        self.error_type = error_type
        self.response = response
        self.disable_retry = disable_retry


class RequestError(YaDiskError):
    """
        Generic exception class for cases when a request could not be sent or
        response could not be received.
    """

    def __init__(self, msg: str = "", disable_retry: bool = False):
        YaDiskError.__init__(self, None, msg, disable_retry=disable_retry)


class YaDiskConnectionError(RequestError):
    """Thrown when a connection error occured."""
    pass


class TooManyRedirectsError(RequestError):
    """Thrown when there were too many redirects."""
    pass


class RequestTimeoutError(RequestError):
    """Thrown when a request timed out."""
    pass


class WrongResourceTypeError(YaDiskError):
    """Thrown when the resource was expected to be of different type (e.g., file instead of directory)."""

    def __init__(self, msg: str = "") -> None:
        YaDiskError.__init__(self, None, msg, None)


class RetriableYaDiskError(YaDiskError):
    """Thrown when there was an error but it would make sense to retry the request."""
    pass


class AsyncOperationFailedError(RetriableYaDiskError):
    """Raised when an asynchronous operation fails"""

    def __init__(self, msg: str = "") -> None:
        YaDiskError.__init__(self, None, msg, None)


class AsyncOperationPollingTimeoutError(YaDiskError):
    """Raised when a polling timeout occured while waiting for an asynchronous operation"""

    def __init__(self, msg: str = "") -> None:
        YaDiskError.__init__(self, None, msg, None)


class UnknownYaDiskError(RetriableYaDiskError):
    """Thrown when the request failed but the response does not contain any error info."""

    def __init__(
        self,
        msg: str= "",
        response: Optional[AnyResponse] = None,
        disable_retry: bool = False
    ) -> None:
        RetriableYaDiskError.__init__(self, None, msg, response, disable_retry=disable_retry)


class BadRequestError(YaDiskError):
    """Thrown when the server returns code 400."""
    pass


class UnauthorizedError(YaDiskError):
    """Thrown when the server returns code 401."""
    pass


class ForbiddenError(YaDiskError):
    """Thrown when the server returns code 403."""
    pass


class NotFoundError(YaDiskError):
    """Thrown when the server returns code 404."""
    pass


class NotAcceptableError(YaDiskError):
    """Thrown when the server returns code 406."""
    pass


class ConflictError(YaDiskError):
    """Thrown when the server returns code 409."""
    pass


class GoneError(YaDiskError):
    """Raised when the server returns code 410."""
    pass


class PayloadTooLargeError(YaDiskError):
    """Thrown when the server returns code 413."""
    pass


class UnsupportedMediaError(YaDiskError):
    """Thrown when the server returns code 415."""
    pass


class LockedError(YaDiskError):
    """Thrown when the server returns code 423."""
    pass


class UploadTrafficLimitExceededError(LockedError):
    """Thrown when upload limit has been exceeded."""
    pass


class TooManyRequestsError(YaDiskError):
    """Thrown when the server returns code 429."""
    pass


class ResourceDownloadLimitExceededError(TooManyRequestsError):
    """Raised when the download limit for a resource is exceeded."""
    pass


class InternalServerError(RetriableYaDiskError):
    """Thrown when the server returns code 500."""
    pass


class BadGatewayError(RetriableYaDiskError):
    """Thrown when the server returns code 502"""
    pass


class UnavailableError(RetriableYaDiskError):
    """Thrown when the server returns code 503."""
    pass


class GatewayTimeoutError(RetriableYaDiskError):
    """Thrown when the server returns code 504"""
    pass


class InsufficientStorageError(YaDiskError):
    """Thrown when the server returns code 507."""
    pass


class PathNotFoundError(NotFoundError):
    """Thrown when the requested path does not exist."""
    pass


class ParentNotFoundError(ConflictError):
    """Thrown by `mkdir`, `upload`, etc. when the parent directory doesn't exist."""
    pass


class PathExistsError(ConflictError):
    """Thrown when the requested path already exists."""
    pass


class DirectoryExistsError(PathExistsError):
    """Thrown when the directory already exists."""
    pass


class FieldValidationError(BadRequestError):
    """Thrown when the request contains fields with invalid data."""
    pass


class ResourceIsLockedError(LockedError):
    """Thrown when the resource is locked by another operation."""
    pass


class MD5DifferError(ConflictError):
    """Thrown when the MD5 hash of the file to be deleted doesn't match with the actual one."""
    pass


class OperationNotFoundError(NotFoundError):
    """Thrown by `get_operation_status()` when the operation doesn't exist."""
    pass


class InvalidResponseError(YaDiskError):
    """Thrown when Yandex.Disk did not return a JSON response or if it's invalid."""
    pass


class AuthorizationPendingError(BadRequestError):
    """Thrown when authorization is currently pending, the application has to wait."""
    pass


class InvalidClientError(BadRequestError):
    """Thrown when an invalid client ID or client secret was provided"""
    pass


class InvalidGrantError(BadRequestError):
    """Thrown when a verification code is expired or invalid"""
    pass


class BadVerificationCodeError(BadRequestError):
    """Thrown when a verification code has invalid format"""
    pass


class UnsupportedTokenTypeError(BadRequestError):
    """Thrown when the specified token cannot be used in a request"""
    pass



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from io import BytesIO
import json
from typing import Any, Optional

from ..exceptions import (
    RequestError, RequestTimeoutError,
    TooManyRedirectsError, YaDiskConnectionError
)

from .._session import Session, Response
from .._typing_compat import Iterator, Tuple, Dict
from ..utils import CaseInsensitiveDict
from ..types import JSON, ConsumeCallback, HTTPMethod, Headers, Payload, TimeoutParameter
from .. import settings

from urllib.parse import urlencode

import pycurl

__all__ = ["PycURLSession"]


def convert_curl_error(error: pycurl.error) -> RequestError:
    code, msg = error.args

    mapping = {pycurl.E_TOO_MANY_REDIRECTS:      TooManyRedirectsError,
               pycurl.E_COULDNT_CONNECT:         YaDiskConnectionError,
               pycurl.E_NO_CONNECTION_AVAILABLE: YaDiskConnectionError,
               pycurl.E_OPERATION_TIMEDOUT:      RequestTimeoutError}

    exc = mapping.get(code) or RequestError
    return exc(msg)


# see PycurlResponse.download() implementation
MAX_RESPONSE_BUFFER_SIZE = 128 * 1024


class PycURLResponse(Response):
    def __init__(self, curl: pycurl.Curl, response: bytes):
        super().__init__()

        self._curl = curl
        self._response = response

        self._update_status()

    def _update_status(self) -> None:
        self.status = self._curl.getinfo(pycurl.RESPONSE_CODE)

    def _perform(self) -> None:
        try:
            self._curl.perform()
        except pycurl.error as e:
            raise convert_curl_error(e) from e

        self._update_status()

    def _perform_rb(self) -> None:
        try:
            self._response = self._curl.perform_rb()
        except pycurl.error as e:
            raise convert_curl_error(e) from e

        self._update_status()

    def json(self) -> JSON:
        if not self.status:
            self._perform_rb()

        return json.loads(self._response)

    def download(self, consume_callback: ConsumeCallback) -> None:
        buffer = BytesIO()

        def write_cb(chunk: bytes) -> int:
            # Write up to `MAX_RESPONSE_BUFFER_SIZE` bytes of data into an in-memory buffer
            # This is a hack to detect bad HTTP status codes to give
            # `consume_callback` an opportunity to check status before writing
            if buffer.tell() < MAX_RESPONSE_BUFFER_SIZE:
                buffer.write(chunk)
                return len(chunk)
            elif buffer.tell():
                buffer.seek(0)

                chunk_from_buffer = buffer.read()
                consume_callback(chunk_from_buffer)

                buffer.seek(0)
                buffer.truncate(0)

            consume_callback(chunk)
            return len(chunk)

        self._curl.setopt(pycurl.WRITEFUNCTION, write_cb)

        self._perform()

        # Write left over data from the buffer
        if buffer.tell():
            buffer.seek(0)
            consume_callback(buffer.read())

    def close(self) -> None:
        self._curl.close()


class IterableReader:
    def __init__(self, iterator: Iterator[bytes]):
        self.iterator = iterator
        self._current_chunk = b""
        self._position_in_chunk = 0

    def read(self, size=-1) -> bytes:
        if size < 0:
            return self.readall()

        data = b""
        while len(data) < size:
            if self._position_in_chunk >= len(self._current_chunk):
                try:
                    self._current_chunk = next(self.iterator)
                except StopIteration:
                    return data

                self._position_in_chunk = 0

            remaining = size - len(data)

            chunk_fragment = self._current_chunk[self._position_in_chunk:self._position_in_chunk + remaining]

            data += chunk_fragment
            self._position_in_chunk += len(chunk_fragment)

        return data

    def readall(self) -> bytes:
        data = b""

        while True:
            if self._position_in_chunk >= len(self._current_chunk):
                try:
                    self._current_chunk = next(self.iterator)
                except StopIteration:
                    return data

                self._position_in_chunk = 0

            chunk_fragment = self._current_chunk[self._position_in_chunk:]

            data += chunk_fragment
            self._position_in_chunk += len(chunk_fragment)


def convert_timeout(timeout: TimeoutParameter) -> Tuple[float, float]:
    if timeout is ...:
        return convert_timeout(settings.DEFAULT_TIMEOUT)

    if isinstance(timeout, tuple):
        connect_timeout, read_timeout = timeout
    else:
        connect_timeout = read_timeout = timeout

    MAX_TIMEOUT = 4294967  # in seconds

    if connect_timeout is None:
        connect_timeout = MAX_TIMEOUT
    elif connect_timeout <= 0.001:
        # If connect_timeout gets rounded down to 0, the default connect
        # timeout would be applied instead by cURL
        connect_timeout = 0.001

    if read_timeout is None:
        # 0 disables LOW_SPEED_TIME
        read_timeout = 0
    elif read_timeout <= 1.0:
        # If read_timeout gets rounded down to 0, the low speed time will be disabled
        # 1 second is the lowest possible timeout
        read_timeout = 1.0

    return connect_timeout, read_timeout

class PycURLSession(Session):
    """
        .. _pycurl: https://pypi.org/project/pycurl

        :any:`Session` implementation using the `pycurl`_ library.

        To pass `pycurl`-specific arguments from :any:`Client` use :code:`curl_options` keyword argument.

        Usage example:

        .. code:: python

           import yadisk
           import pycurl

           with yadisk.Client(..., session="pycurl") as client:
               client.get_meta(
                   "/my_file.txt",
                   n_retries=5,
                   curl_options={
                       pycurl.MAX_SEND_SPEED_LARGE: 5 * 1024**2,
                       pycurl.MAX_RECV_SPEED_LARGE: 5 * 1024**2,
                       pycurl.PROXY: "http://localhost:12345",
                       pycurl.MAXREDIRS: 15
                   }
                )
    """

    def __init__(self) -> None:
        self._share = pycurl.CurlShare()

        self._share.setopt(pycurl.SH_SHARE, pycurl.LOCK_DATA_CONNECT)
        self._share.setopt(pycurl.SH_SHARE, pycurl.LOCK_DATA_DNS)
        self._share.setopt(pycurl.SH_SHARE, pycurl.LOCK_DATA_SSL_SESSION)

    def send_request(
        self,
        method: HTTPMethod,
        url: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Payload] = None,
        headers: Optional[Headers] = None,
        stream: bool = False,
        curl_options: Optional[Dict[int, Any]] = None,
        **kwargs
    ) -> Response:
        curl_headers = CaseInsensitiveDict({"connection": "keep-alive"})
        curl_headers.update(headers or {})

        if params:
            url = url + "?" + urlencode(params)

        curl = pycurl.Curl()
        curl.setopt(pycurl.NOSIGNAL, True)
        curl.setopt(pycurl.FOLLOWLOCATION, True)
        curl.setopt(pycurl.URL, url)
        curl.setopt(pycurl.SHARE, self._share)

        if "timeout" in kwargs:
            connect_timeout, read_timeout = convert_timeout(kwargs["timeout"])

            curl.setopt(pycurl.CONNECTTIMEOUT_MS, int(connect_timeout * 1000))
            curl.setopt(pycurl.LOW_SPEED_TIME, int(read_timeout))
            curl.setopt(pycurl.LOW_SPEED_LIMIT, 64)

        curl.setopt(pycurl.HTTPHEADER, [f"{k}:{v}" for k, v in curl_headers.items() if k and v])


        if curl_options is not None:
            for option, value in curl_options.items():
                curl.setopt(option, value)

        if isinstance(data, bytes):
            data = BytesIO(data)

        uploading_file = False

        if data is not None:
            curl.setopt(pycurl.UPLOAD, True)
            uploading_file = True

            if isinstance(data, Iterator):
                curl_data = IterableReader(data)
            else:
                curl_data = data

            curl.setopt(pycurl.READDATA, curl_data)

        curl.setopt(pycurl.CUSTOMREQUEST, method)

        if not stream or uploading_file:
            try:
                response = curl.perform_rb()
            except pycurl.error as e:
                raise convert_curl_error(e) from e
        else:
            response = b""

        return PycURLResponse(curl, response)

    def close(self) -> None:
        self._share.close()



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from ._yadisk_object import YaDiskObject
from .._common import str_or_error, int_or_error

from typing import Any, Optional

__all__ = ["DeviceCodeObject", "TokenObject", "TokenRevokeStatusObject"]


class TokenObject(YaDiskObject):
    """
        Token object.

        :param token: `dict` or `None`
        :param yadisk: :any:`YaDisk` or `None`, `YaDisk` object

        :ivar access_token: `str`, token string
        :ivar refresh_token: `str`, the refresh-token
        :ivar token_type: `str`, type of the token
        :ivar expires_in: `int`, amount of time before the token expires
        :ivar scope: `str`, list of rights requested by the application,
                     returned only if the token has a smaller set of rights
                     than requested
    """

    access_token: Optional[str]
    refresh_token: Optional[str]
    token_type: Optional[str]
    expires_in: Optional[int]
    scope: Optional[str]

    def __init__(self, token: Optional[dict] = None, yadisk: Optional[Any] = None):
        YaDiskObject.__init__(
            self,
            {"access_token":  str_or_error,
             "refresh_token": str_or_error,
             "token_type":    str_or_error,
             "expires_in":    int_or_error,
             "scope":         str_or_error},
            yadisk)

        self.import_fields(token)


class TokenRevokeStatusObject(YaDiskObject):
    """
        Result of token revocation request.

        :param token_revoke_status: `dict` or `None`
        :param yadisk: :any:`YaDisk` or `None`, `YaDisk` object

        :ivar status: `str`, status of the operation
    """

    status: Optional[str]

    def __init__(self,
                 token_revoke_status: Optional[dict] = None,
                 yadisk: Optional[Any] = None):
        YaDiskObject.__init__(self, {"status": str_or_error}, yadisk)

        self.import_fields(token_revoke_status)


class DeviceCodeObject(YaDiskObject):
    """
        Result of :any:`Client.get_device_code()` / :any:`AsyncClient.get_device_code()`.

        :param device_code_object: `dict` or `None`
        :param yadisk: :any:`YaDisk` or `None`, `YaDisk` object

        :ivar device_code: `str`, device code that can be used for obtaining the token
        :ivar user_code: `str`, code that the user should enter on the OAuth page
        :ivar verification_url: `str`, URL of the OAuth page where user is
                                expected to enter the :code:`user_code`
        :ivar interval: `int`, the minimum interval (in seconds) with which the
                        app must request an OAuth token. If requests come more
                        often, Yandex OAuth may respond with an error
        :ivar expires_in: `int`, amount of time before the codes expire
    """

    device_code:      Optional[str]
    user_code:        Optional[str]
    verification_url: Optional[str]
    interval:         Optional[int]
    expires_in:       Optional[int]

    def __init__(self,
                 device_code_object: Optional[dict] = None,
                 yadisk: Optional[Any] = None):
        YaDiskObject.__init__(
            self,
            {
                "device_code":      str_or_error,
                "user_code":        str_or_error,
                "verification_url": str_or_error,
                "interval":         int_or_error,
                "expires_in":       int_or_error
            },
            yadisk)

        self.import_fields(device_code_object)



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from ._yadisk_object import YaDiskObject
from ._link_object import LinkObject
from .._common import str_or_error
from ..types import OperationStatus

from typing import Any, Optional

__all__ = [
    "AsyncOperationLinkObject",
    "OperationLinkObject",
    "OperationStatusObject",
    "SyncOperationLinkObject"
]


class OperationStatusObject(YaDiskObject):
    """
        Operation status object.

        :param operation_status: `dict` or `None`
        :param yadisk: :any:`YaDisk` or `None`, `YaDisk` object

        :ivar status: `str`, status of the operation
    """

    status: OperationStatus

    def __init__(self,
                 operation_status: Optional[dict] = None,
                 yadisk: Optional[Any] = None):
        YaDiskObject.__init__(
            self,
            {"status": str_or_error},
            yadisk)

        self.import_fields(operation_status)


class OperationLinkObject(LinkObject):
    """
        Operation link object.

        :param link: `dict` or `None`
        :param yadisk: :any:`Client`/:any:`AsyncClient` or `None`, `YaDisk` object

        :ivar href: `str`, link URL
        :ivar method: `str`, HTTP method
        :ivar templated: `bool`, tells whether the URL is templated
    """

    pass


class SyncOperationLinkObject(OperationLinkObject):
    """
        Operation link object.

        :param link: `dict` or `None`
        :param yadisk: :any:`Client` or `None`, `YaDisk` object

        :ivar href: `str`, link URL
        :ivar method: `str`, HTTP method
        :ivar templated: `bool`, tells whether the URL is templated
    """

    def get_status(self, **kwargs) -> OperationStatus:
        """
            Get operation status.

            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises OperationNotFoundError: requested operation was not found

            :returns: `str`, :code:`"in-progress"` indicates that the operation
                      is currently running, :code:`"success"` indicates that
                      the operation was successful, :code:`"failed"` means that
                      the operation failed
        """

        if self._yadisk is None:
            raise ValueError("This object is not bound to a YaDisk instance")

        if self.href is None:
            raise ValueError("OperationLinkObject has no link")

        return self._yadisk.get_operation_status(self.href, **kwargs)

    def wait(self, **kwargs) -> None:
        """
            Wait until an operation is completed. If the operation fails, an
            exception is raised. Waiting is performed by calling :any:`time.sleep`.

            :param poll_interval: `float`, interval in seconds between subsequent operation status queries
            :param poll_timeout: `float` or `None`, total polling timeout (`None` means no timeout),
                                 if this timeout is exceeded, an exception is raised
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises OperationNotFoundError: requested operation was not found
            :raises AsyncOperationFailedError: requested operation failed
            :raises AsyncOperationPollingTimeoutError: requested operation did not
                                                       complete in specified time
                                                       (when `poll_timeout` is not `None`)
        """

        if self._yadisk is None:
            raise ValueError("This object is not bound to a YaDisk instance")

        if self.href is None:
            raise ValueError("OperationLinkObject has no link")

        return self._yadisk.wait_for_operation(self.href, **kwargs)


class AsyncOperationLinkObject(OperationLinkObject):
    """
        Operation link object.

        :param link: `dict` or `None`
        :param yadisk: :any:`AsyncClient` or `None`, `YaDisk` object

        :ivar href: `str`, link URL
        :ivar method: `str`, HTTP method
        :ivar templated: `bool`, tells whether the URL is templated
    """

    async def get_status(self, **kwargs) -> OperationStatus:
        """
            Get operation status.

            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param aiohttp_args: `dict`, additional parameters for :any:`AIOHTTPSession`
            :param httpx_args: `dict`, additional parameters for :any:`AsyncHTTPXSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises OperationNotFoundError: requested operation was not found

            :returns: `str`, :code:`"in-progress"` indicates that the operation
                      is currently running, :code:`"success"` indicates that
                      the operation was successful, :code:`"failed"` means that
                      the operation failed
        """

        if self._yadisk is None:
            raise ValueError("This object is not bound to a YaDisk instance")

        if self.href is None:
            raise ValueError("OperationLinkObject has no link")

        return await self._yadisk.get_operation_status(self.href, **kwargs)

    async def wait(self, **kwargs) -> None:
        """
            Wait until an operation is completed. If the operation fails, an
            exception is raised. Waiting is performed by calling :any:`asyncio.sleep`.

            :param poll_interval: `float`, interval in seconds between subsequent operation status queries
            :param poll_timeout: `float` or `None`, total polling timeout (`None` means no timeout),
                                 if this timeout is exceeded, an exception is raised
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param aiohttp_args: `dict`, additional parameters for :any:`AIOHTTPSession`
            :param httpx_args: `dict`, additional parameters for :any:`AsyncHTTPXSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises OperationNotFoundError: requested operation was not found
            :raises AsyncOperationFailedError: requested operation failed
            :raises AsyncOperationPollingTimeoutError: requested operation did not
                                                       complete in specified time
                                                       (when `poll_timeout` is not `None`)
        """

        if self._yadisk is None:
            raise ValueError("This object is not bound to a YaDisk instance")

        if self.href is None:
            raise ValueError("OperationLinkObject has no link")

        return await self._yadisk.wait_for_operation(self.href, **kwargs)



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from typing import Any, Optional
from .._typing_compat import Callable, Iterator

__all__ = ["YaDiskObject"]


class YaDiskObject:
    """
        Base class for all objects mirroring the ones returned by Yandex.Disk REST API.
        It must have a fixed number of fields, each field must have a type.
        It also supports subscripting and access of fields through the . operator.

        :param field_types: `dict` or `None`
        :param yadisk: :any:`YaDisk` or `None`, `YaDisk` object
    """

    FIELD_TYPES: dict
    FIELDS: dict
    ALIASES: dict
    _yadisk: Optional[Any]

    def __init__(self,
                 field_types: Optional[dict] = None,
                 yadisk: Optional[Any] = None):
        if field_types is None:
            field_types = {}

        self.FIELD_TYPES = {}
        self.FIELDS = {}
        self.ALIASES = {}
        self.set_field_types(field_types)

        self._yadisk = yadisk

    def set_field_types(self, field_types: dict) -> None:
        """
            Set the field types of the object

            :param field_types: `dict`, where keys are the field names and values are types (or factories)
        """

        self.FIELD_TYPES = field_types

        for field in field_types.keys():
            self[field] = None

    def set_field_type(self, field: str, type: Callable) -> None:
        """
            Set field type.

            :param field: `str`
            :param type: type or factory
        """

        self.FIELD_TYPES[field] = type
        self[field] = None

    def set_alias(self, alias: str, name: str) -> None:
        """
            Set an alias.

            :param alias: `str`, alias to add
            :param name: `str`, field name
        """

        self.ALIASES[alias] = name

    def remove_alias(self, alias: str) -> None:
        """
            Remove an alias.

            :param alias: `str`
        """

        self.ALIASES.pop(alias)

    def remove_field(self, field: str) -> None:
        """
            Remove field.

            :param field: `str`
        """

        self.FIELDS.pop(field)
        self.FIELD_TYPES.pop(field)

    def import_fields(self, source_dict: Optional[dict]) -> None:
        """
            Set all the fields of the object to the values in `source_dict`.
            All the other fields are ignored

            :param source_dict: `dict` or `None` (nothing will be done in that case)
        """

        if source_dict is not None:
            for field in self.FIELDS:
                try:
                    self[field] = source_dict[field]
                except KeyError:
                    pass

            for alias, field in self.ALIASES.items():
                try:
                    self[field] = source_dict[alias]
                except KeyError:
                    pass

    def __setattr__(self, attr: str, value: Any) -> None:
        if attr in ("FIELDS", "FIELD_TYPES", "ALIASES", "_yadisk"):
            self.__dict__[attr] = value
            return

        attr = self.ALIASES.get(attr, attr)

        if attr not in self.FIELD_TYPES:
            raise AttributeError("Unknown attribute: %r" % (attr,))

        datatype = self.FIELD_TYPES[attr]
        self.FIELDS[attr] = datatype(value) if value is not None else None

    def __getattr__(self, attr: str) -> Any:
        attr = self.ALIASES.get(attr, attr)

        if attr not in self.FIELD_TYPES:
            raise AttributeError("Unknown attribute: %r" % (attr,))

        return self.FIELDS[attr]

    def __getitem__(self, key: str) -> Any:
        return self.FIELDS[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.__setattr__(key, value)

    def __delitem__(self, key: str) -> None:
        del self.FIELDS[key]

    def __iter__(self) -> Iterator[dict]:
        return iter(self.FIELDS)

    def __len__(self) -> int:
        return len(self.FIELDS)

    def __repr__(self) -> str:
        return "<%s%r>" % (self.__class__.__name__, self.FIELDS)

    def _repr_pretty_(self, p, cycle: bool) -> None:
        """IPython pretty-print implementation."""

        if cycle:
            p.text(f"<{self.__class__.__name__}{'{...}'}>")
        else:
            if not self.FIELDS:
                p.text(f"<{self.__class__.__name__}{'{}'}>")
                return

            with p.group(4, f"<{self.__class__.__name__}{'{'}", "})>"):
                p.breakable()

                for idx, (k, v) in enumerate(self.FIELDS.items()):
                    if idx:
                        p.text(",")
                        p.breakable()

                    p.text(repr(k))
                    p.text(": ")
                    p.pretty(v)

    def field(self, name: str) -> Any:
        """
            Get value of field `name`, guarantee it's not :code:`None` or
            raise a :any:`ValueError`.

            :param name: `str`, name of the field

            :raises ValueError: value of the given field is :code:`None`

            :returns: field's value
        """

        value = self.__getattr__(name)

        if value is not None:
            return value

        raise ValueError(f"field {repr(name)} is None")

    def __matmul__(self, name: str) -> Any:
        """
            The :code:`@` operator. Same as :any:`YaDiskObject.field()`.
            Can be used like this:

            .. code:: python

                # if embedded or embedded.total turn out to be None, we'll get a ValueError
                file_count = client.get_meta("/some_folder") @ "embedded" @ "total"
                print(f"/some_folder contains {file_count} files")

            :param name: `str`, name of the field

            :raises ValueError: value of the given field is :code:`None`

            :returns: field's value
        """

        return self.field(name)



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from .api_request import APIRequest
from ..objects import (
    PublicResourcesListObject, SyncPublicResourcesListObject, AsyncPublicResourcesListObject,
    TrashResourceObject, SyncTrashResourceObject, AsyncTrashResourceObject,
    FilesResourceListObject, SyncFilesResourceListObject, AsyncFilesResourceListObject,
    LastUploadedResourceListObject, SyncLastUploadedResourceListObject, AsyncLastUploadedResourceListObject,
    ResourceObject, SyncResourceObject, AsyncResourceObject, ResourceUploadLinkObject,
    PublicResourceObject, SyncPublicResourceObject, AsyncPublicResourceObject,
    OperationLinkObject, SyncOperationLinkObject, AsyncOperationLinkObject,
    ResourceLinkObject, SyncResourceLinkObject, AsyncResourceLinkObject, ResourceDownloadLinkObject
)

from .._common import is_operation_link, ensure_path_has_schema
from ..exceptions import InvalidResponseError

from .._typing_compat import Iterable, Dict, List
from typing import Optional, Union, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ..types import AnySession, AnyClient, JSON

__all__ = [
    "CopyRequest",
    "DeleteRequest",
    "DeleteTrashRequest",
    "FilesRequest",
    "GetDownloadLinkRequest",
    "GetMetaRequest",
    "GetPublicDownloadLinkRequest",
    "GetPublicMetaRequest",
    "GetPublicResourcesRequest",
    "GetTrashRequest",
    "GetUploadLinkRequest",
    "LastUploadedRequest",
    "MkdirRequest",
    "MoveRequest",
    "PatchRequest",
    "PublishRequest",
    "RestoreTrashRequest",
    "SaveToDiskRequest",
    "UnpublishRequest",
    "UploadURLRequest",
]

Fields = Iterable[str]


def _substitute_keys(keys: Iterable[str], sub_map: Dict[str, str]) -> List[str]:
    return [".".join(sub_map.get(f, f) for f in k.split(".")) for k in keys]


class GetPublicResourcesRequest(APIRequest):
    """
        A request to get a list of public resources.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param offset: offset from the beginning of the list
        :param limit: maximum number of elements in the list
        :param preview_size: size of the file preview
        :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
        :param type: filter based on type of resources ("file" or "dir")
        :param fields: list of keys to be included in the response

        :returns: :any:`PublicResourcesListObject`
    """

    method = "GET"
    path = "/v1/disk/resources/public"

    def __init__(
        self,
        session:      "AnySession",
        offset:       int                     = 0,
        limit:        int                     = 20,
        preview_size: Optional[str]           = None,
        preview_crop: Optional[bool]          = None,
        type:         Optional[str]           = None,
        fields:       Optional[Iterable[str]] = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        self.params["offset"] = offset
        self.params["limit"] = limit

        if preview_size is not None:
            self.params["preview_size"] = preview_size

        if preview_crop is not None:
            self.params["preview_crop"] = preview_crop

        if type is not None:
            self.params["type"] = type

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> PublicResourcesListObject:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        if yadisk is None or yadisk.synchronous:
            return SyncPublicResourcesListObject(js, yadisk)
        else:
            return AsyncPublicResourcesListObject(js, yadisk)


class UnpublishRequest(APIRequest):
    """
        A request to make a public resource private.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param path: path to the resource to be unpublished
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceLinkObject`
    """

    method = "PUT"
    path = "/v1/disk/resources/unpublish"

    def __init__(
        self,
        session:  "AnySession",
        path:     str,
        fields:   Optional[Iterable[str]] = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        self.params["path"] = ensure_path_has_schema(path)

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> ResourceLinkObject:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        if yadisk is None or yadisk.synchronous:
            return SyncResourceLinkObject(js, yadisk)
        else:
            return AsyncResourceLinkObject(js, yadisk)


class GetDownloadLinkRequest(APIRequest):
    """
        A request to get a download link to a resource.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param path: path to the resource to be downloaded
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceDownloadLinkObject`
    """

    method = "GET"
    path = "/v1/disk/resources/download"

    def __init__(
        self,
        session:  "AnySession",
        path:     str,
        fields:   Optional[Iterable[str]] = None,
        **kwargs
    ):
        APIRequest.__init__(self, session, **kwargs)

        self.params["path"] = ensure_path_has_schema(path)

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> ResourceDownloadLinkObject:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        return ResourceDownloadLinkObject(js, yadisk)


class GetTrashRequest(APIRequest):
    """
        A request to get meta-information about a trash resource.

        :param path: path to the trash resource
        :param limit: number of children resources to be included in the response
        :param offset: number of children resources to be skipped in the response
        :param preview_size: size of the file preview
        :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
        :param sort: `str`, field to be used as a key to sort children resources
        :param fields: list of keys to be included in the response

        :returns: :any:`TrashResourceObject`
    """

    method = "GET"
    path = "/v1/disk/trash/resources"

    def __init__(
        self,
        session:      "AnySession",
        path:         str,
        offset:       int                     = 0,
        limit:        int                     = 20,
        sort:         Optional[str]           = None,
        preview_size: Optional[str]           = None,
        preview_crop: Optional[bool]          = None,
        fields:       Optional[Iterable[str]] = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        self.params["path"] = ensure_path_has_schema(path, "trash")
        self.params["offset"] = offset
        self.params["limit"] = limit

        if sort is not None:
            self.params["sort"] = sort

        if preview_size is not None:
            self.params["preview_size"] = preview_size

        if preview_crop is not None:
            self.params["preview_crop"] = preview_crop

        if fields is not None:
            sub_map = {"embedded": "_embedded"}

            self.params["fields"] = ",".join(_substitute_keys(fields, sub_map))

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> TrashResourceObject:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        if yadisk is None or yadisk.synchronous:
            return SyncTrashResourceObject(js, yadisk)
        else:
            return AsyncTrashResourceObject(js, yadisk)


class RestoreTrashRequest(APIRequest):
    """
        A request to restore trash.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param path: path to the trash resource to be restored
        :param dst_path: destination path
        :param force_async: forces the operation to be executed asynchronously
        :param overwrite: `bool`, determines whether the destination can be overwritten
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceLinkObject` or :any:`OperationLinkObject`
    """

    method = "PUT"
    path = "/v1/disk/trash/resources/restore"
    success_codes = {201, 202}

    def __init__(
        self,
        session:     "AnySession",
        path:        str,
        dst_path:    Optional[str]           = None,
        force_async: bool                    = False,
        overwrite:   bool                    = False,
        fields:      Optional[Iterable[str]] = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        self.params["path"] = ensure_path_has_schema(path, "trash")
        self.params["overwrite"] = "true" if overwrite else "false"
        self.params["force_async"] = "true" if force_async else "false"

        if dst_path is not None:
            self.params["name"] = dst_path

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> Union[OperationLinkObject, ResourceLinkObject]:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON", disable_retry=True)

        if is_operation_link(js.get("href", "")):
            if yadisk is None or yadisk.synchronous:
                return SyncOperationLinkObject(js, yadisk)
            else:
                return AsyncOperationLinkObject(js, yadisk)
        elif self.params.get("force_async") == "true":
            raise InvalidResponseError(
                "Yandex.Disk did not return an operation link, despite force_async=true",
                disable_retry=True
            )

        if yadisk is None or yadisk.synchronous:
            return SyncResourceLinkObject(js, yadisk)
        else:
            return AsyncResourceLinkObject(js, yadisk)


class DeleteTrashRequest(APIRequest):
    """
        A request to delete a trash resource.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param path: path to the trash resource to be deleted
        :param force_async: forces the operation to be executed asynchronously
        :param fields: list of keys to be included in the response

        :returns: :any:`OperationLinkObject` or `None`
    """

    method = "DELETE"
    path = "/v1/disk/trash/resources"
    success_codes = {202, 204}

    def __init__(
        self,
        session:     "AnySession",
        path:        Optional[str]           = None,
        force_async: bool                    = False,
        fields:      Optional[Iterable[str]] = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        if path is not None:
            self.params["path"] = ensure_path_has_schema(path, "trash")

        self.params["force_async"] = "true" if force_async else "false"

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> Optional[OperationLinkObject]:
        if js is not None:
            if not isinstance(js, dict):
                raise InvalidResponseError("Yandex.Disk returned invalid JSON", disable_retry=True)

            if yadisk is None or yadisk.synchronous:
                return SyncOperationLinkObject(js, yadisk)
            else:
                return AsyncOperationLinkObject(js, yadisk)
        elif self.params.get("force_async") == "true":
            raise InvalidResponseError(
                "Yandex.Disk did not return an operation link, despite force_async=true",
                disable_retry=True
            )

        return None


class LastUploadedRequest(APIRequest):
    """
        A request to get the list of latest uploaded files sorted by upload date.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param limit: maximum number of elements in the list
        :param media_type: type of files to include in the list
        :param preview_size: size of the file preview
        :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
        :param fields: list of keys to be included in the response

        :returns: :any:`LastUploadedResourceListObject`
    """

    method = "GET"
    path = "/v1/disk/resources/last-uploaded"

    def __init__(
        self,
        session: "AnySession",
        limit:        int                                 = 20,
        media_type:   Optional[Union[str, Iterable[str]]] = None,
        preview_size: Optional[str]                       = None,
        preview_crop: Optional[bool]                      = None,
        fields:       Optional[Iterable[str]]             = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        self.params["limit"] = limit

        if media_type is not None:
            if not isinstance(media_type, Iterable):
                raise TypeError("media_type should be a string or an iterable")

            if isinstance(media_type, str):
                self.params["media_type"] = media_type
            else:
                self.params["media_type"] = ",".join(media_type)

        if preview_size is not None:
            self.params["preview_size"] = preview_size

        if preview_crop is not None:
            self.params["preview_crop"] = preview_crop

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> LastUploadedResourceListObject:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        if yadisk is None or yadisk.synchronous:
            return SyncLastUploadedResourceListObject(js, yadisk)
        else:
            return AsyncLastUploadedResourceListObject(js, yadisk)


class CopyRequest(APIRequest):
    """
        A request to copy a file or a directory.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param src_path: source path
        :param dst_path: destination path
        :param overwrite: if `True` the destination path can be overwritten,
                          otherwise, an error will be raised
        :param force_async: forces the operation to be executed asynchronously
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceLinkObject` or :any:`OperationLinkObject`
    """

    method = "POST"
    path = "/v1/disk/resources/copy"
    success_codes = {201, 202}

    def __init__(
        self,
        session:     "AnySession",
        src_path:    str,
        dst_path:    str,
        overwrite:   bool             = False,
        force_async: bool             = False,
        fields:      Optional[Fields] = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        self.params["from"] = ensure_path_has_schema(src_path)
        self.params["path"] = ensure_path_has_schema(dst_path)
        self.params["overwrite"] = "true" if overwrite else "false"
        self.params["force_async"] = "true" if force_async else "false"

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> Union[OperationLinkObject, ResourceLinkObject]:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON", disable_retry=True)

        if is_operation_link(js.get("href", "")):
            if yadisk is None or yadisk.synchronous:
                return SyncOperationLinkObject(js, yadisk)
            else:
                return AsyncOperationLinkObject(js, yadisk)
        elif self.params.get("force_async") == "true":
            raise InvalidResponseError(
                "Yandex.Disk did not return an operation link, despite force_async=true",
                disable_retry=True
            )

        if yadisk is None or yadisk.synchronous:
            return SyncResourceLinkObject(js, yadisk)
        else:
            return AsyncResourceLinkObject(js, yadisk)


class GetMetaRequest(APIRequest):
    """
        A request to get meta-information about a resource.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param path: path to the resource
        :param limit: number of children resources to be included in the response
        :param offset: number of children resources to be skipped in the response
        :param preview_size: size of the file preview
        :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
        :param sort: `str`, field to be used as a key to sort children resources
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceObject`
    """

    method = "GET"
    path = "/v1/disk/resources"

    def __init__(
        self,
        session: "AnySession",
        path:         str,
        limit:        Optional[int]    = None,
        offset:       Optional[int]    = None,
        preview_size: Optional[str]    = None,
        preview_crop: Optional[bool]   = None,
        sort:         Optional[str]    = None,
        fields:       Optional[Fields] = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        self.params["path"] = ensure_path_has_schema(path)

        if limit is not None:
            self.params["limit"] = limit

        if offset is not None:
            self.params["offset"] = offset

        if preview_size is not None:
            self.params["preview_size"] = preview_size

        if preview_crop is not None:
            self.params["preview_crop"] = "true" if preview_crop else "false"

        if sort is not None:
            self.params["sort"] = sort

        if fields is not None:
            sub_map = {"embedded": "_embedded"}

            self.params["fields"] = ",".join(_substitute_keys(fields, sub_map))

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> ResourceObject:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        if yadisk is None or yadisk.synchronous:
            return SyncResourceObject(js, yadisk)
        else:
            return AsyncResourceObject(js, yadisk)


class GetUploadLinkRequest(APIRequest):
    """
        A request to get an upload link.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param path: path to be uploaded at
        :param overwrite: `bool`, determines whether to overwrite the destination
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceUploadLinkObject`
    """

    method = "GET"
    path = "/v1/disk/resources/upload"

    def __init__(
        self,
        session:   "AnySession",
        path:      str,
        overwrite: bool             = False,
        fields:    Optional[Fields] = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        self.params["path"] = ensure_path_has_schema(path)
        self.params["overwrite"] = "true" if overwrite else "false"

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> ResourceUploadLinkObject:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        return ResourceUploadLinkObject(js, yadisk)


class MkdirRequest(APIRequest):
    """
        A request to create a new directory.

        :param path: path to the directory to be created
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceLinkObject`
    """

    method = "PUT"
    path = "/v1/disk/resources"
    success_codes = {201}

    def __init__(
        self,
        session:  "AnySession",
        path:     str,
        fields:   Optional[Fields] = None,
        **kwargs
    ):
        APIRequest.__init__(self, session, **kwargs)

        self.params["path"] = ensure_path_has_schema(path)

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> ResourceLinkObject:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON", disable_retry=True)

        if yadisk is None or yadisk.synchronous:
            return SyncResourceLinkObject(js, yadisk)
        else:
            return AsyncResourceLinkObject(js, yadisk)


class PublishRequest(APIRequest):
    """
        A request to make a resource public.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param path: path to the resource to be published
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceLinkObject`
    """

    method = "PUT"
    path = "/v1/disk/resources/publish"

    def __init__(
        self,
        session:  "AnySession",
        path:     str,
        fields:   Optional[Fields] = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        self.params["path"] = ensure_path_has_schema(path)

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> ResourceLinkObject:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        if yadisk is None or yadisk.synchronous:
            return SyncResourceLinkObject(js, yadisk)
        else:
            return AsyncResourceLinkObject(js, yadisk)


class UploadURLRequest(APIRequest):
    """
        A request to upload a file from URL.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param url: source URL
        :param path: destination path
        :param disable_redirects: `bool`, forbid redirects
        :param fields: list of keys to be included in the response

        :returns: :any:`OperationLinkObject`
    """

    method = "POST"
    path = "/v1/disk/resources/upload"
    success_codes = {202}

    def __init__(
        self,
        session:           "AnySession",
        url:               str,
        path:              str,
        disable_redirects: bool             = False,
        fields:            Optional[Fields] = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        self.params["url"] = url
        self.params["path"] = ensure_path_has_schema(path)
        self.params["disable_redirects"] = "true" if disable_redirects else "false"

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> OperationLinkObject:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON", disable_retry=True)

        if yadisk is None or yadisk.synchronous:
            return SyncOperationLinkObject(js, yadisk)
        else:
            return AsyncOperationLinkObject(js, yadisk)


class DeleteRequest(APIRequest):
    """
        A request to delete a file or a directory.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param path: path to the resource to be removed
        :param permanently: if `True`, the resource will be removed permanently,
                            otherwise, it will be just moved to the trash
        :param force_async: forces the operation to be executed asynchronously
        :param md5: `str`, MD5 hash of the file to remove
        :param fields: list of keys to be included in the response

        :returns: :any:`OperationLinkObject` or `None`
    """

    method = "DELETE"
    path = "/v1/disk/resources"
    success_codes = {202, 204}

    def __init__(
        self,
        session:     "AnySession",
        path:        str,
        permanently: bool             = False,
        md5:         Optional[str]    = None,
        force_async: bool             = False,
        fields:      Optional[Fields] = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        self.params["path"] = ensure_path_has_schema(path)
        self.params["permanently"] = "true" if permanently else "false"
        self.params["force_async"] = "true" if force_async else "false"

        if md5 is not None:
            self.params["md5"] = md5

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> Optional[OperationLinkObject]:
        if isinstance(js, dict):
            if yadisk is None or yadisk.synchronous:
                return SyncOperationLinkObject(js, yadisk)
            else:
                return AsyncOperationLinkObject(js, yadisk)
        elif js is not None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON", disable_retry=True)
        elif self.params.get("force_async") == "true":
            raise InvalidResponseError(
                "Yandex.Disk did not return an operation link, despite force_async=true",
                disable_retry=True
            )

        return None


class SaveToDiskRequest(APIRequest):
    """
        A request to save a public resource to the disk.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param public_key: public key or public URL of the resource
        :param name: filename of the saved resource
        :param path: path to the copied resource in the public folder
        :param save_path: path to the destination directory (downloads directory by default)
        :param force_async: forces the operation to be executed asynchronously
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceLinkObject` or :any:`OperationLinkObject`
    """

    method = "POST"
    path = "/v1/disk/public/resources/save-to-disk"
    success_codes = {201, 202}

    def __init__(
        self,
        session:     "AnySession",
        public_key:  str,
        name:        Optional[str]    = None,
        path:        Optional[str]    = None,
        save_path:   Optional[str]    = None,
        force_async: bool             = False,
        fields:      Optional[Fields] = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        self.params["public_key"] = public_key

        if name is not None:
            self.params["name"] = name

        if path is not None:
            self.params["path"] = path

        if save_path is not None:
            self.params["save_path"] = ensure_path_has_schema(save_path)

        self.params["force_async"] = "true" if force_async else "false"

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> Union[OperationLinkObject, ResourceLinkObject]:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON", disable_retry=True)

        if is_operation_link(js.get("href", "")):
            if yadisk is None or yadisk.synchronous:
                return SyncOperationLinkObject(js, yadisk)
            else:
                return AsyncOperationLinkObject(js, yadisk)
        elif self.params.get("force_async") == "true":
            raise InvalidResponseError(
                "Yandex.Disk did not return an operation link, despite force_async=true",
                disable_retry=True
            )

        if yadisk is None or yadisk.synchronous:
            return SyncResourceLinkObject(js, yadisk)
        else:
            return AsyncResourceLinkObject(js, yadisk)


class GetPublicMetaRequest(APIRequest):
    """
        A request to get meta-information about a public resource.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param public_key: public key or public URL of the resource
        :param path: relative path to a resource in a public folder.
                     By specifying the key of the published folder in `public_key`,
                     you can request metainformation for any resource in the folder.
        :param offset: offset from the beginning of the list of nested resources
        :param limit: maximum number of nested elements to be included in the list
        :param sort: `str`, field to be used as a key to sort children resources
        :param preview_size: file preview size
        :param preview_crop: `bool`, allow preview crop
        :param fields: list of keys to be included in the response

        :returns: :any:`PublicResourceObject`
    """

    method = "GET"
    path = "/v1/disk/public/resources"

    def __init__(
        self,
        session:      "AnySession",
        public_key:   str,
        offset:       int              = 0,
        limit:        int              = 20,
        path:         Optional[str]    = None,
        sort:         Optional[str]    = None,
        preview_size: Optional[str]    = None,
        preview_crop: Optional[bool]   = None,
        fields:       Optional[Fields] = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        self.params["public_key"] = public_key
        self.params["offset"] = offset
        self.params["limit"] = limit

        if path is not None:
            self.params["path"] = path

        if sort is not None:
            self.params["sort"] = sort

        if preview_size is not None:
            self.params["preview_size"] = preview_size

        if preview_crop is not None:
            self.params["preview_crop"] = preview_crop

        if fields is not None:
            sub_map = {"embedded": "_embedded",
                       "view_count": "views_count"}

            self.params["fields"] = ",".join(_substitute_keys(fields, sub_map))

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> PublicResourceObject:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        if yadisk is None or yadisk.synchronous:
            return SyncPublicResourceObject(js, yadisk)
        else:
            return AsyncPublicResourceObject(js, yadisk)


class GetPublicDownloadLinkRequest(APIRequest):
    """
        A request to get a download link for a public resource.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param public_key: public key or public URL of the resource
        :param path: relative path to the resource within the public folder
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceDownloadLinkObject`
    """

    method = "GET"
    path = "/v1/disk/public/resources/download"

    def __init__(
        self,
        session:    "AnySession",
        public_key: str,
        path:       Optional[str]    = None,
        fields:     Optional[Fields] = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        self.params["public_key"] = public_key

        if path is not None:
            self.params["path"] = path

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> ResourceDownloadLinkObject:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        return ResourceDownloadLinkObject(js, yadisk)


class MoveRequest(APIRequest):
    """
        A request to move a resource.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param src_path: source path to be moved
        :param dst_path: destination path
        :param force_async: forces the operation to be executed asynchronously
        :param overwrite: `bool`, determines whether to overwrite the destination
        :param fields: list of keys to be included in the response

        :returns: :any:`OperationLinkObject` or :any:`ResourceLinkObject`
    """

    method = "POST"
    path = "/v1/disk/resources/move"
    success_codes = {201, 202}

    def __init__(
        self,
        session:      "AnySession",
        src_path:    str,
        dst_path:    str,
        force_async: bool             = False,
        overwrite:   bool             = False,
        fields:      Optional[Fields] = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        self.params["from"] = ensure_path_has_schema(src_path)
        self.params["path"] = ensure_path_has_schema(dst_path)
        self.params["overwrite"] = "true" if overwrite else "false"
        self.params["force_async"] = "true" if force_async else "false"

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> Union[OperationLinkObject, ResourceLinkObject]:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON", disable_retry=True)

        if is_operation_link(js.get("href", "")):
            if yadisk is None or yadisk.synchronous:
                return SyncOperationLinkObject(js, yadisk)
            else:
                return AsyncOperationLinkObject(js, yadisk)
        elif self.params.get("force_async") == "true":
            raise InvalidResponseError(
                "Yandex.Disk did not return an operation link, despite force_async=true",
                disable_retry=True
            )

        if yadisk is None or yadisk.synchronous:
            return SyncResourceLinkObject(js, yadisk)
        else:
            return AsyncResourceLinkObject(js, yadisk)


class FilesRequest(APIRequest):
    """
        A request to get a flat list of all files (that doesn't include directories).

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param offset: offset from the beginning of the list
        :param limit: number of list elements to be included
        :param media_type: type of files to include in the list
        :param sort: `str`, field to be used as a key to sort children resources
        :param preview_size: size of the file preview
        :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
        :param fields: list of keys to be included in the response

        :returns: :any:`FilesResourceListObject`
    """

    method = "GET"
    path = "/v1/disk/resources/files"

    def __init__(
        self,
        session:      "AnySession",
        offset:       int                                 = 0,
        limit:        int                                 = 20,
        media_type:   Optional[Union[str, Iterable[str]]] = None,
        preview_size: Optional[str]                       = None,
        preview_crop: Optional[bool]                      = None,
        sort:         Optional[str]                       = None,
        fields:       Optional[Fields]                    = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        self.params["offset"] = offset
        self.params["limit"] = limit

        if media_type is not None:
            if not isinstance(media_type, Iterable):
                raise TypeError("media_type should be a string or an iterable")

            if isinstance(media_type, str):
                self.params["media_type"] = media_type
            else:
                self.params["media_type"] = ",".join(media_type)

        if preview_size is not None:
            self.params["preview_size"] = preview_size

        if preview_crop is not None:
            self.params["preview_crop"] = preview_crop

        if sort is not None:
            self.params["sort"] = sort

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> FilesResourceListObject:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        if yadisk is None or yadisk.synchronous:
            return SyncFilesResourceListObject(js, yadisk)
        else:
            return AsyncFilesResourceListObject(js, yadisk)


class PatchRequest(APIRequest):
    """
        A request to update custom properties of a resource.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param path: path to the resource
        :param properties: `dict`, custom properties to update
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceObject`
    """

    method = "PATCH"
    path = "/v1/disk/resources"
    content_type = "application/json"

    def __init__(
        self,
        session:    "AnySession",
        path:       str,
        properties: dict,
        fields:     Optional[Fields] = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        self.params["path"] = ensure_path_has_schema(path)
        self.data = {"custom_properties": properties}

        if fields is not None:
            sub_map = {"embedded": "_embedded"}

            self.params["fields"] = ",".join(_substitute_keys(fields, sub_map))

    def process_json(
        self,
        js: "JSON",
        yadisk: Optional["AnyClient"] = None,
        **kwargs
    ) -> ResourceObject:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        if yadisk is None or yadisk.synchronous:
            return SyncResourceObject(js, yadisk)
        else:
            return AsyncResourceObject(js, yadisk)



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from ._yadisk_object import YaDiskObject

from typing import Any, Optional

from .._common import str_or_error, bool_or_error

__all__ = ["LinkObject"]


class LinkObject(YaDiskObject):
    """
        Link object.

        :param link: `dict` or `None`
        :param yadisk: :any:`Client`/:any:`AsyncClient` or `None`, `YaDisk` object

        :ivar href: `str`, link URL
        :ivar method: `str`, HTTP method
        :ivar templated: `bool`, tells whether the URL is templated
    """

    href: Optional[str]
    method: Optional[str]
    templated: Optional[bool]

    def __init__(self,
                 link: Optional[dict] = None,
                 yadisk: Optional[Any] = None):
        YaDiskObject.__init__(
            self,
            {"href":      str_or_error,
             "method":    str_or_error,
             "templated": bool_or_error},
            yadisk)

        self.import_fields(link)
